#include "pogohash.h"

//#include "pogohash.conf.41.4"
//#include "pogohash.conf.43.3"
//#include "pogohash.conf.45.0"
  #include "pogohash.conf.47.1"

/**
https://www.trustwave.com/Resources/SpiderLabs-Blog/Jailbreak-Detection-Methods/
 **/

uc_engine *uc;
int        uc_ver;
csh        cs;
uint8_t   *code_buffer;
int        initialized = 0;

const int HASHSEED = 0x61247FBF;

uint8_t neon[] = { 
                   0x4f, 0xf4, 0x70, 0x00,  // mov.w r0, #0xf00000
                   0x01, 0xee, 0x50, 0x0f,  // mcr p15, #0x0, r0, c1, c0, #0x2
                   0xbf, 0xf3, 0x6f, 0x8f,  // isb sy 
                   0x4f, 0xf0, 0x80, 0x43,  // mov.w r3, #0x40000000
                   0xe8, 0xee, 0x10, 0x3a   // vmsr fpexc, r3
                 }; 
uint8_t stub[] = { 
                   0x90, 0x47,              // blx r2
                   0x00, 0xbf               // nop
                 };
uint32_t HashFunctionAddr = 0;

uc_hook hhc_ins;
void POGO_hook(uc_engine *uc, uint64_t address, uint32_t size, void *user_data)
{
#ifdef TRACE_DEBUG
  if (initialized == 1)
  {
    static uint32_t regs[16]     = { 0 };
    static uint32_t regs_tmp[16] = { 0 };
    static int      ins_counter;
    static int      address_lo;
    static int      address_hi;
    static int      address_last = 0x2000;

    // instruction counter
    if (address == POGO_FUNC)
    {
      address_lo = address;
      address_hi = address;
      ins_counter = 1;
    }
    else
    if (address > 0x2002)
    {
      if (address < address_lo) address_lo = address;
      if (address > address_hi) address_hi = address;
      ins_counter++;
    }

    if ((((uint64_t)address - (uint64_t)address_last) < 0) ||
        (((uint64_t)address - (uint64_t)address_last) > 4))
    {
      printf("DETECT: BRANCH\n");
    }

    printf("---\n");
    printf("address: 0x%08lx\n", address);
    printf("  ");
    {
      uint32_t reg;
      int      i;

      // we must map the capstone register names against unicorn
      for (i=0; i<16; i++)
      {
        switch (i)
        {
          case  9: uc_reg_read(uc, UC_ARM_REG_SB, &reg); printf(" SB"); break;
          case 10: uc_reg_read(uc, UC_ARM_REG_SL, &reg); printf(" SL"); break;
          case 11: uc_reg_read(uc, UC_ARM_REG_FP, &reg); printf(" FP"); break;
          case 12: uc_reg_read(uc, UC_ARM_REG_IP, &reg); printf(" IP"); break;
          case 13: uc_reg_read(uc, UC_ARM_REG_SP, &reg); printf(" SP"); break;
          case 14: uc_reg_read(uc, UC_ARM_REG_LR, &reg); printf(" LR"); break;
          case 15:                        reg = address; printf(" PC"); break; 
          default: uc_reg_read(uc, UC_ARM_REG_R0+i, &reg); printf("R%02d", i);
                   break;
        }
        if (regs[i] != reg) printf("*"); else printf(":");
        printf(" 0x%08x ", reg); 
        regs[i]     = reg;
        regs_tmp[i] = reg;
  
        if ((i % 4) == 3) printf("\n  ");
      }

      // print flags
      {
        uint32_t flags;
        int n, z, c, v;

        // BUG: flags do not seem to get updated, investigate later
        flags = 0;
        uc_reg_read(uc, UC_ARM_REG_CPSR, &flags);
        n = ((flags & 0x80000000) != 0);
        z = ((flags & 0x40000000) != 0);
        c = ((flags & 0x20000000) != 0);
        v = ((flags & 0x10000000) != 0);

        printf("FLAGS N%d Z%d C%d V%d\n", n, z, c, v);
        printf("       ");
        printf(" eq=%d", (z==1) ? 1 : 0);
        printf(" ne=%d", (z==0) ? 1 : 0);
        printf(" cs=%d", (c==1) ? 1 : 0);
        printf(" cc=%d", (c==0) ? 1 : 0);
        printf(" vs=%d", (v==1) ? 1 : 0);
        printf(" vc=%d", (v==0) ? 1 : 0);
        printf(" hi=%d", (c==1) && (z==0) ? 1 : 0);
        printf(" ls=%d", (c==0) || (z==1) ? 1 : 0);
        printf(" ge=%d", (n==v) ? 1 : 0);
        printf(" lt=%d", (n!=v) ? 1 : 0);
        printf(" gt=%d", (z==0) && (n==v) ? 1 : 0);
        printf(" le=%d", (z==1) || (n!=v) ? 1 : 0);
      }
    }
    printf("\n");

    // display the instruction (hack)
    {
      cs_insn* insn;
      int      i, cnt;

#define MOV_INST_FLEX_SPAN 8

      static int      movlo[16]     = { 0 };
      static uint16_t movlo_val[16] = { 0 };
      static int      movhi[16]     = { 0 };
      static uint16_t movhi_val[16] = { 0 };
      static int      mov_flex[16]  = { 0 };

      cnt = cs_disasm(cs, &code_buffer[address], 10, address, 1, &insn);
      if (cnt > 0)
      {
        printf("opcodes: ");
        for (i=0; i<insn->size; i++) printf("%02x ", insn->bytes[i]);
        for (i=insn->size; i<4; i++) printf("   ");

        printf(":: %s\t%s", insn->mnemonic, insn->op_str);

        // we want to detect when 32-bit values are loaded in to registers
        if (strncmp(insn->mnemonic, "mov", 3) == 0)
        {
          char     *_p, *p;
          uint16_t  val;
          int       r;

          // flag if we find a movX instruction
          r = -1;

          // high 16-bit: movt
          if (strncmp(insn->mnemonic, "movt", 4) == 0) 
          {
            // we must map the capstone register names against unicorn
            p = _p = insn->op_str;
                 if (strncmp(p, "sb", 2) == 0) r = 9;
            else if (strncmp(p, "sl", 2) == 0) r = 10;
            else if (strncmp(p, "fp", 2) == 0) r = 11;
            else if (strncmp(p, "ip", 2) == 0) r = 12;
            else if (strncmp(p, "sp", 2) == 0) r = 13;
            else if (strncmp(p, "lr", 2) == 0) r = 14;
            else if (strncmp(p, "pc", 2) == 0) r = 15;
            else 
            {
              p++;
              r = strtol(p, &_p, 10);
            }
            p = strstr(_p, "#"); p++;
            val = strtol(p, &_p, 16);

            // apply to temp register (we must validate before showing)
            regs_tmp[r] = (regs_tmp[r] & 0x0000ffff) | (val << 16);

            // this is for tracking the changes
            movhi[r] = 1;
            movhi_val[r] = val;
            mov_flex[r] = 0; // reset our flexibility factor
          }
          else

          // low 16-bit: movw/movs
          if ((strncmp(insn->mnemonic, "movw", 4) == 0) ||
              (strncmp(insn->mnemonic, "movs", 4) == 0)) 
          {
            // we must map the capstone register names against unicorn
            p = _p = insn->op_str;
                 if (strncmp(p, "sb", 2) == 0) r = 9;
            else if (strncmp(p, "sl", 2) == 0) r = 10;
            else if (strncmp(p, "fp", 2) == 0) r = 11;
            else if (strncmp(p, "ip", 2) == 0) r = 12;
            else if (strncmp(p, "sp", 2) == 0) r = 13;
            else if (strncmp(p, "lr", 2) == 0) r = 14;
            else if (strncmp(p, "pc", 2) == 0) r = 15;
            else 
            {
              p++;
              r = strtol(p, &_p, 10);
            }
            p = strstr(_p, "#"); p++;
            val = strtol(p, &_p, 16);

            // apply to temp register (we must validate before showing)
            if (insn->mnemonic[3] == 's')
              regs_tmp[r] = (regs_tmp[r] & 0xffffff00) | val; 
            else
              regs_tmp[r] = (regs_tmp[r] & 0xffff0000) | val; 

            // this is for tracking the changes
            movlo[r] = 1;
            movlo_val[r] = val;
            mov_flex[r] = 0; // reset our flexibility factor
          }

          // did we have a movX and have both (relative)
          if ((r != -1) && (movhi[r] == 1) && (movlo[r] == 1))
          {
            uint32_t val = ((uint32_t)movhi_val[r] << 16) | movlo_val[r];

            // make sure temp register matches the value (avoid corruption)
            if (val == regs_tmp[r]) 
            {
              printf("\n");
              printf("DETECT: LOAD R%02d 0x%08x @[%08x]", 
                     r, val, (uint32_t)address);
            }
          }
        }

        // we have a lot of conditions like add rx, pc - can filter
        if ((strncmp(insn->mnemonic, "add", 3) == 0) && 
            (strstr(insn->op_str, ", pc") != NULL))
        {
          char *_p, *p;
          int   r;

          // we must map the capstone register names against unicorn
          p = _p = insn->op_str;
               if (strncmp(p, "sb", 2) == 0) r = 9;
          else if (strncmp(p, "sl", 2) == 0) r = 10;
          else if (strncmp(p, "fp", 2) == 0) r = 11;
          else if (strncmp(p, "ip", 2) == 0) r = 12;
          else if (strncmp(p, "sp", 2) == 0) r = 13;
          else if (strncmp(p, "lr", 2) == 0) r = 14;
          else if (strncmp(p, "pc", 2) == 0) r = 15;
          else 
          {
            p++;
            r = strtol(p, &_p, 10);
          }

          // we want to log this
          {
            uint32_t val = ((uint32_t)movhi_val[r] << 16) | movlo_val[r];
            printf("\n");
            printf("DETECT: ADDPC R%02d 0x%08x @[%08x]", 
                    r, val, (uint32_t)address);
          }
        }

        for (i=0; i<16; i++)
        {
          // last mov instruction on register too long ago? 
          mov_flex[i] = mov_flex[i] + 1;
          if (mov_flex[i] > (MOV_INST_FLEX_SPAN-1))
          {
            movhi[i] = 0; 
            movlo[i] = 0; 
          }
        }

        cs_free(insn, cnt);
      }
      else
        printf("INVALID INSTRUCTION");
    }
    printf("\n");

    // instruction counter
    if (address == POGO_FUNC_END)
    {
      printf("\n");
      printf("address_lo: 0x%08lx ", (uint64_t)address_lo);
      printf("address_hi: 0x%08lx ", (uint64_t)address_hi);
      printf("instructions: %d\n", ins_counter);
    }

    address_last = address;
  }

  // make sure everything is written; just in case of a crash
  fflush(stdout);
#endif
}

uc_hook hhc_mem_rw;
bool POGO_hook_mem(uc_engine *uc, uc_mem_type type, uint64_t address, int size, int64_t value, void *user_data)
{
#ifdef TRACE_DEBUG
  // we ignore anything above 0xd0000000 (heap, stack)
  if (address < 0xd0000000) 
  {   
    switch (size)
    {
      case 4:
           if (type == UC_MEM_READ)
             printf("[mem-r] len:%d at 0x%08lx val: 0x%08x\n", 
                    size, address, *((uint32_t *)&code_buffer[address]));
           else
           if (type == UC_MEM_WRITE)
             printf("[mem-w] len:%d at 0x%08lx val: 0x%08x\n", 
                    size, address, (uint32_t)value);
           break;
    }
  }
  else
  {
    switch (size)
    {
      case 4:
           if (type == UC_MEM_READ)
             printf("[mem-r] len:%d at 0x%08lx\n", 
                    size, address);
           else
           if (type == UC_MEM_WRITE)
             printf("[mem-w] len:%d at 0x%08lx val: 0x%08x\n", 
                    size, address, (uint32_t)value);
           break;
    }
  }
#endif
  return false;
}

uc_err POGO_createmap(uint64_t address, uint32_t size)
{
  uint32_t chunk = address % 0x1000;
  uint64_t taddr = address - chunk;
  uint32_t tsize = ((0xFFF + chunk + size) / 0x1000) * 0x1000;

  uc_err err = uc_mem_map(uc, taddr, tsize, UC_PROT_ALL);
  if (err)
  {
    printf("Failed on uc_mem_map() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

uc_err POGO_memmap(uint64_t address, uint8_t* buffer, uint32_t size)
{
  uc_err err = uc_mem_write(uc, address, buffer, size);
  if (err)
  {
    printf("Failed on uc_mem_write() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

uc_err POGO_memget(uint64_t address, uint8_t* buffer, uint32_t size)
{
  uc_err err = uc_mem_read(uc, address, buffer, size);
  if (err)
  {
    printf("Failed on uc_mem_read() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

int POGO_init()
{
  unsigned int sp;
  uc_err err;
  FILE* fp;
  long size, i;
  int exportdata, exportaddr, cnt;
 
  // dynamically load shared library
#ifdef DYNLOAD
  if (!uc_dyn_load(NULL, 0))
  {
    printf("Error dynamically loading shared library.\n");
    printf("Please check that unicorn.dll/unicorn.so is available as well as\n");
    printf("any other dependent dll/so files.\n");
    printf("The easiest way is to place them in the same directory as this app.\n");
    return 1;
  }
#endif

  // Initialize emulator in ARM 32bit thumb mode
  err = uc_open(UC_ARCH_ARM, UC_MODE_THUMB, &uc);
  if (err)
  {
    printf("Failed on uc_open() with error returned: %u\n", err);
    return err;
  }

  // we need the version
  uc_ver = uc_version(NULL, NULL);
  printf("uc_ver = 0x%04x\n", uc_ver);

  // hook all instructions by having @begin > @end
  uc_hook_add(uc, &hhc_ins, UC_HOOK_CODE, (void *)POGO_hook, NULL, (uint64_t)1, (uint64_t)0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  uc_hook_add(uc, &hhc_mem_rw, UC_HOOK_MEM_READ | UC_HOOK_MEM_WRITE, POGO_hook_mem, NULL, 1, 0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }

  // heap and stack allocation
  POGO_createmap(0xE0000000, 0x8000);
  sp = 0xE0000000 + 0x7000; // stack grows down in heap, allow for underflow

  // stack
  uc_reg_write(uc, UC_ARM_REG_SP, &sp);

  // binary mapping
  fp = fopen(POGO_FILENAME, "rb");
  size = POGO_BIN_SIZE; 
  code_buffer = malloc(size);
  fseek(fp, 0, SEEK_SET);
  fread(code_buffer, size, 1, fp);
  fclose(fp);

  // memory prepare
  POGO_createmap(POGO_BIN_OFFSET, POGO_BIN_MAX - POGO_BIN_OFFSET);

#if (POGO_PATCH_VER == 41)
  // specific for pokemon.41.4
  {
    char nop[2] = { 0x00, 0xbf }; // NOP 
    memmove(code_buffer + 0x01BE9D9E, nop, sizeof(nop)); 
  }
#endif
#if (POGO_PATCH_VER == 45)
  // specific for pokemon.45.0
  {
    char nop[2] = { 0x00, 0xbf }; // NOP 
    memmove(code_buffer + 0x01B1769C, nop, sizeof(nop)); 
    memmove(code_buffer + 0x01B1769E, nop, sizeof(nop)); 
    memmove(code_buffer + 0x01B17CD4, nop, sizeof(nop)); 
    memmove(code_buffer + 0x01B17CD6, nop, sizeof(nop)); 
  }
  // specific for pokemon.45.0: end
#endif
#if (POGO_PATCH_VER == 47) 
  // specific for pokemon.47.1
  {
    char nop[2]        = { 0x00, 0xbf };             // nop 
    char movr0_0[2]    = { 0x00, 0x20 };             // mov r0, #0 
    char movr0_neg1[4] = { 0x4f, 0xf0, 0xff, 0x30 }; // mov r0, #-1
 
    // address: 0x01b1207e opcodes: 90 f2 c6 c6 :: blx     #0x29a2e0c
/**
 1b1207e:       f290 c6c6       blx     0x29a2e0c
 1b12082:       f104 0098       add.w   r0, r4, #152    ; 0x98
 1b12086:       900a            str     r0, [sp, #40]   ; 0x28
 **/
    memmove(code_buffer + 0x01B1207E, nop, sizeof(nop)); 
    memmove(code_buffer + 0x01B12080, nop, sizeof(nop)); 

    // address: 0x01dc9ac6 opcodes; 80 47       :: blx     r0 
/**
 1dc9ac6:       4780            blx     r0
 1dc9ac8:       b138            cbz     r0, 0x1dc9ada
 1dc9aca:       2100            movs    r1, #0
 1dc9acc:       f1b0 3fff       cmp.w   r0, #4294967295 ; 0xffffffff
...
 1dc9ada:       2000            movs    r0, #0
 1dc9adc:       47a0            blx     r4      
 **/
    // fork() should return -1 on failure
    memmove(code_buffer + 0x01DC9AC6, movr0_neg1, sizeof(movr0_neg1)); 

    // address: 0x01dc9af4 opcodes: 88 47       :: blx     r1
/**
 1dc9af4:       4788            blx     r1
 1dc9af6:       2800            cmp     r0, #0
 1dc9af8:       bf18            it      ne
 1dc9afa:       2001            movne   r0, #1
 **/
    // system() should return 0 on failure
    memmove(code_buffer + 0x01DC9AF4, movr0_0, sizeof(movr0_0)); 

    // address: 0x01b12e62 opcodes: 8f f2 e4 c7 :: blx     #0x29a2e2c
/**
 1b12e62:       f28f c7e4       blx     0x29a2e2c
 1b12e66:       989b            ldr     r0, [sp, #620]  ; 0x26c
 **/
    memmove(code_buffer + 0x01B12E62, nop, sizeof(nop)); 
    memmove(code_buffer + 0x01B12E64, nop, sizeof(nop)); 
  }
  // specific for pokemon.47.1: end
#endif


  // neon support enable 
  POGO_createmap(0x1000, 0x1000);
  POGO_memmap(0x1000, neon, sizeof(neon));
#ifdef TRACE_DEBUG
  memmove(code_buffer + 0x1000, neon, sizeof(neon)); 
#endif

  // start code calling hash func
  POGO_createmap(0x2000, 0x1000);
  POGO_memmap(0x2000, stub, sizeof(stub));
#ifdef TRACE_DEBUG
  memmove(code_buffer + 0x2000, stub, sizeof(stub)); 
#endif

  // export address fill with any valid address 
  exportdata = POGO_EXP_DATA;
  exportaddr = POGO_EXP_ADDR;
  cnt = 0;
  for (i = 0; i < POGO_EXP_CNT; i++)
  {
    int value = *(int*)(code_buffer + exportaddr + (i * 4));
    if (value == 0)
    {
      memmove(code_buffer + exportaddr + (i * 4), &exportdata, 4); 
      cnt++;
    }
  }
  printf("%d patches\n", cnt);

  // all segment
  POGO_memmap(POGO_BIN_OFFSET, 
              code_buffer + POGO_BIN_OFFSET, 
              POGO_BIN_SIZE - POGO_BIN_OFFSET);

  // enable neon
  {   
    uint32_t addr;
    uint32_t addr_end;

    addr     = 0x1000;
    addr_end = 0x1000 + sizeof(neon);
    if (uc_ver > 0x09) addr |= 1; // required after 0.9
    err = uc_emu_start(uc, addr, addr_end, 0, 0);
    if (err)
    {
      printf("Failed on uc_emu_start() with error returned %u: %s\n",
        err, uc_strerror(err));
      return -1;
    }
  }

#ifdef TRACE_DEBUG
  // initialize the disassembler
  {
    cs_err err = cs_open(CS_ARCH_ARM, CS_MODE_THUMB, &cs);
    if (err)
    {
      printf("Failed on cs_open() with error returned %u: %s\n",
        err, cs_strerror(err));
      return -1;
    }

    cs_option(cs, CS_OPT_DETAIL, CS_OPT_OFF);
  }  
#endif

  HashFunctionAddr = POGO_FUNC;
  initialized = 1;

  return 0;
}

void POGO_quit()
{
#ifdef TRACE_DEBUG
  cs_close(&cs);
#endif

  // free resources
  uc_close(uc);

  free(code_buffer);

  // dynamically free shared library
#ifdef DYNLOAD
  uc_dyn_free();
#endif
}

uint64_t Hash(uint8_t* buffer, int size)
{
  uint32_t r0 = 0xE0000000;  // buffer
  uint32_t r1 = size;        // size
  uint32_t r2 = HashFunctionAddr + 1;
  uint64_t ret;
  uint32_t addr;
  uint32_t addr_end;
  uc_err err;

  uc_reg_write(uc, UC_ARM_REG_R0, &r0);
  uc_reg_write(uc, UC_ARM_REG_R1, &r1);
  uc_reg_write(uc, UC_ARM_REG_R2, &r2);

  // copy the buffer to the memory at 0xe000000
  POGO_memmap(r0, buffer, size);

  // execute code
  addr     = 0x2000;
  addr_end = 0x2000 + sizeof(stub);
  if (uc_ver > 0x09) addr |= 1; // required after 0.9
  err = uc_emu_start(uc, addr, addr_end, 0, 0);
  if (err)
  {
    printf("Failed on uc_emu_start() with error returned %u: %s\n",
      err, uc_strerror(err));
    return -1;
  }

  // read the buffer from memory at 0xe000000
  POGO_memget(0xE0000000, buffer, size);

  uc_reg_read(uc, UC_ARM_REG_R0, &r0);
  uc_reg_read(uc, UC_ARM_REG_R1, &r1);

  ret = (unsigned int)r1;
  ret = (ret << 32) | (unsigned int)r0;
  return (int64_t)ret;
}

int64_t Hash64(char* buffer, int size, int salt)
{
  int64_t ret;
  uint8_t* newbuff = (uint8_t *)malloc(size + 4);
  newbuff[0] = (salt >> 24) & 0xFF;
  newbuff[1] = (salt >> 16) & 0xFF;
  newbuff[2] = (salt >> 8) & 0xFF;
  newbuff[3] = (salt >> 0) & 0xFF;
  memcpy(newbuff + 4, buffer, size);
  ret = Hash(newbuff, size + 4);
  free(newbuff);

  return ret;
}

int Hash64to32(int64_t hash)
{
  return (unsigned int)(hash >> 0) ^ (unsigned int)(hash >> 32);
}

int Hash32(char* buffer, int size)
{
  int64_t ret = Hash64(buffer, size, HASHSEED);
  return Hash64to32(ret);
}
